"""pandas utilities."""
import io
import logging
import textwrap
from typing import Any

import pandas as pd

from wpdatautil.timeit import Timer

log = logging.getLogger(__name__)


def csv_text_to_df(text: str) -> pd.DataFrame:
    """Return the dataframe constructed from the given CSV table text."""
    return pd.read_csv(io.StringIO(textwrap.dedent(text)))


def dedup_df(df: pd.DataFrame, *, df_description: str = "dataframe", log_if_rate_kept_le: float = 0, **kwargs: Any) -> float:
    """Return the rate kept after deduplicating the dataframe in-place.

    The reason for using this function is that it meaningfully logs the outcome.

    `kwargs` are forwarded to `df.drop_duplicates`.
    """
    len_pre_drop = len(df)
    log.debug(f"Dropping duplicate rows from {len_pre_drop:,} rows of the {df_description} using: {kwargs}.")
    timer = Timer()
    assert "inplace" not in kwargs
    df.drop_duplicates(**kwargs, inplace=True)
    len_post_drop = len(df)
    len_dropped = len_pre_drop - len_post_drop
    rate_dropped = (len_dropped / len_pre_drop) if len_pre_drop else float("nan")
    rate_kept = (len_post_drop / len_pre_drop) if len_pre_drop else float("nan")
    logger = log.info if (rate_kept <= log_if_rate_kept_le) else log.debug
    logger(f"Dropped {len_dropped:,} ({rate_dropped:.2%}) duplicate rows in {timer} from the {df_description}, resulting in {len_post_drop:,} ({rate_kept:.2%}) rows.")
    return rate_kept


def deindex_df(df: pd.DataFrame, *, index: Any, df_description: str = "dataframe", log_if_rate_kept_le: float = 0) -> float:
    """Return the rate kept after dropping index labels from the dataframe in-place.

    The reason for using this function is that it meaningfully logs the outcome.
    """
    len_pre_drop = len(df)
    log.debug(f"Dropping rows matching index labels from {len_pre_drop:,} rows of the {df_description}.")
    timer = Timer()
    df.drop(index=index, inplace=True)
    len_post_drop = len(df)
    len_dropped = len_pre_drop - len_post_drop
    rate_dropped = (len_dropped / len_pre_drop) if len_pre_drop else float("nan")
    rate_kept = (len_post_drop / len_pre_drop) if len_pre_drop else float("nan")
    logger = log.info if (rate_kept <= log_if_rate_kept_le) else log.debug
    logger(f"Dropped {len_dropped:,} ({rate_dropped:.2%}) rows matching index labels in {timer} from the {df_description}, resulting in {len_post_drop:,} ({rate_kept:.2%}) rows.")
    return rate_kept


def denull_df(df: pd.DataFrame, *, df_description: str = "dataframe", criteria_description: str = "criteria", log_if_rate_kept_le: float = 0, **kwargs: Any) -> float:
    """Return the rate kept after dropping rows having null values from the dataframe in-place.

    The reason for using this function is that it meaningfully logs the outcome.

    `kwargs` are forwarded to `df.dropna`.
    """
    len_pre_drop = len(df)
    log.debug(f"Dropping rows having {criteria_description} from {len_pre_drop:,} rows of the {df_description} using: {kwargs}")
    timer = Timer()
    assert "inplace" not in kwargs
    df.dropna(**kwargs, inplace=True)
    len_post_drop = len(df)
    len_dropped = len_pre_drop - len_post_drop
    rate_dropped = (len_dropped / len_pre_drop) if len_pre_drop else float("nan")
    rate_kept = (len_post_drop / len_pre_drop) if len_pre_drop else float("nan")
    logger = log.info if (rate_kept <= log_if_rate_kept_le) else log.debug
    logger(
        f"Dropped {len_dropped:,} ({rate_dropped:.2%}) rows in {timer} having {criteria_description} from the {df_description}, "
        f"resulting in {len_post_drop:,} ({rate_kept:.2%}) rows."
    )
    return rate_kept


def set_pandas_display_options() -> None:
    """Set pandas display options."""
    # Ref: https://stackoverflow.com/a/52432757/
    display = pd.options.display

    display.max_columns = 1000
    display.max_rows = 1000
    display.max_colwidth = 199
    display.width = None
    # display.precision = 2  # set as needed



"""pathlib utilities."""
from pathlib import Path


def ensure_parent_dir(path: Path) -> None:
    """Ensure that the parent directory of the given path exists."""
    path.parent.mkdir(parents=True, exist_ok=True)



"""pandas base64 utilities."""
from hashlib import shake_128

import pandas as pd
from pybase64 import urlsafe_b64encode  # Benchmarked to be slightly faster than Python's base64 module over millions of rows.


def series_to_str_id(pd_series: pd.Series, /) -> str:
    """Return a reproducibly random string ID of 128 bits which is 16 bytes which is 24 urlsafe base64 characters.

    Usage pattern:
    >>> assert len(df.drop_duplicates()) <= 2 ** 64
    >>> df['xyz_id'] = df.apply(series_to_str_id, axis="columns").astype("string")

    In practice it is more efficient to apply the function only to unique rows, and then merge the results.
    """
    # Notes:
    #  A cache is intentionally not used because the function is expected to be called exactly once for each key.
    #  A 50% probability of a collision is at 2**64 inputs.
    #  A digest of length 32 must not be used because it will produce 44 base64 characters which could subsequently undesirably get mutated by Datavant Link.

    # Verbose version:
    # seed = str(tuple(pd_series)).encode()  # str(tuple(pd_series)) (without any index) is required for reproducibility.
    # hasher = shake_128(seed)
    # digest = hasher.digest(16)  # pylint: disable=too-many-function-args
    # id_ = urlsafe_b64encode(digest).decode()
    # return id_

    return urlsafe_b64encode(shake_128(str(tuple(pd_series)).encode()).digest(16)).decode()  # pylint: disable=too-many-function-args
